package rabinizer.exec;
import java.util.*;

import rabinizer.formulas.*;

public class GeneralizedRabinStatesAcceptanceCondition {
	
	 public List<Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>, List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>>> condition = null;    
         
         public Map<Integer,Map<Integer,Integer>> indexToInt = null;
	
		
	GeneralizedRabinStatesAcceptanceCondition(KDFAStates kdfa){
		condition = new ArrayList<Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>, List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>>>(kdfa.rCondition.size());
		
		for(int i=0; i<kdfa.rCondition.size();i++){
			Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>> empty = new HashSet<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>();
			List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>> empty2 = new ArrayList<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>();
			for(int j=0;j< kdfa.rCondition.get(i).right.size();j++){
				//Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> e=kdfa.rCondition.get(i).right.get(j);
				Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>> empty3 = new HashSet<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>();
				empty2.add(j,empty3);
			}
			Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>,List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>> emptyTuple = new Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>, List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>>(empty,empty2);
			condition.add(i,emptyTuple);
					
		}
		
		for (Tuple<Tuple<Formula, KState>, List<List<Integer>>> state : kdfa.states) {
			for(int i=0; i<kdfa.rCondition.size();i++){
				if(state.right.get(i).get(0)==1){
					condition.get(i).left.add(state);
				}
				for(int j=0;j< kdfa.rCondition.get(i).right.size();j++){
					//Misc.verboseln("xxxxxxxxx"+i+j+state);
					//Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> e=kdfa.rCondition.get(i).right.get(j);
					if(state.right.get(kdfa.rCondition.size()+i).get(j)==1){
						condition.get(i).right.get(j).add(state);
						//Misc.verboseln("addking");
					}
				}
				
			}
			
		}
		
		
		
		
		
		
			
	}
		
		
	
	public String toString() {
        int rpc = 0;
        String r = "Rabin pair count: "+condition.size()+"\n";
        for (Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>, List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>> pair : condition) {
            r += "Pair "+(++rpc)+":\n";
        	r += pairToString(pair);
        }
        return r;
    }
    
	
    private String pairToString(Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>, List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>> pair) {
    	String r = "[\n";
        for (Tuple<Tuple<Formula, KState>, List<List<Integer>>> avoidRabinState : pair.left)
            r += " "+avoidRabinState+"\n";
        r += "]\n[\n";
        for (Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>> reachRabinState : pair.right) {
        	r += " [ \n";
        	for (Tuple<Tuple<Formula, KState>, List<List<Integer>>> reachGenRabinState : reachRabinState)
        		r += "  "+reachGenRabinState+"\n";
        	r += " ] \n";
        }
        r += "]\n";
        return r;
    }
    
    public GenAccCondPrism toBitSet (Map<Tuple<Tuple<Formula, KState>, List<List<Integer>>>, Integer> ordering) {
    	GenAccCondPrism cond = new GenAccCondPrism();
    	for (Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>, List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>> pair : condition) {
    		BitSet fin = stateSetToBitSet(pair.left, ordering);
    		ArrayList<BitSet> infList = new ArrayList<BitSet>(pair.right.size());
    		for (Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>> infSet : pair.right){
    			BitSet inf = stateSetToBitSet(infSet,ordering);
    			infList.add(inf);
    		}
    		cond.addAccCond(fin, infList);
    	}
    	return cond;
    }
    
    private BitSet stateSetToBitSet (Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>> stateSet, Map<Tuple<Tuple<Formula, KState>, List<List<Integer>>>, Integer> ordering) {
    	BitSet bs = new BitSet(stateSet.size());
    	for (Tuple<Tuple<Formula, KState>, List<List<Integer>>> state : stateSet)
    		bs.set(ordering.get(state));
    	return bs;
    }
    
    public int size() {
    	return condition.size();
    }
    
    public String toNumericalString() {
        String r = "";
        Integer rpc = 0;
        Integer rpc2 = 0;
        String id;
        indexToInt = new HashMap<Integer, Map<Integer, Integer>>(condition.size());
        for (Tuple<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>, List<Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>>>> pair : condition) {
            if (rpc2 > 0)
                r += " | ";
            indexToInt.put(rpc2, new HashMap<Integer,Integer>());
            r += "(Fin(" + rpc + ")";
            indexToInt.get(rpc2).put(0, rpc);
            rpc++;
            
            int i=1;
            for (Set<Tuple<Tuple<Formula, KState>, List<List<Integer>>>> reachSet : pair.right) {
                r += "&Inf(" + rpc + ")";
                indexToInt.get(rpc2).put(i, rpc);
                rpc++;
                i++;
            }
            rpc2++;
            r += ")";
        }
        return rpc + " " + r;
    }
    
  
}
