package rabinizer.exec;

import java.util.ArrayList;
import java.lang.*;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Map.Entry;

//import jdd.util.Array;
import rabinizer.formulas.Formula;

/**
 * 
 * @author Zuzana Komarkova
 *
 */

public class KRabinAutomaton {
	
    public Set<Tuple<Tuple<Formula, KState>, List<Integer>>> states = null;
    public Tuple<Tuple<Formula, KState>, List<Integer>> initialState = null;
    public Map<Tuple<Tuple<Tuple<Formula, KState>,List<Integer>>, Tuple<Tuple<Formula, KState>,List<Integer>>>, ValuationSet> edges = null;

    public List<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> condition = null;
    //public Set<Tuple<Map<Tuple<Tuple<Tuple<Formula, KState>,List<Integer>>, Tuple<Tuple<Formula, KState>,List<Integer>>>, ValuationSet>, Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> conditionRabin = null;
    //public List< Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>> conRabHelp= null;
    public Map<Formula, Formula> displayLabels = null;
    
    public Map<Tuple<Tuple<Formula, KState>, List<Integer>>,Integer> statesToNumbers=null;
    public Integer currentNumber;
    public String str;
    
    KRabinAutomaton(KDFA dfaA, GeneralizedRabinAcceptanceCondition grac) {
    	Misc.verboseln("Rabin Automaton");
    	
    	displayLabels = dfaA.displayLabels;
    	
    	//fix the order of the pairs and the order of the reach eventually
    	condition = new ArrayList<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>>();
    	for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair : grac.condition){
    		List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>> reachEventually = new ArrayList<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>(pair.right.size());
    		reachEventually.addAll(pair.right);
    		condition.add(new Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>(pair.left, reachEventually));
    	}
    	
    	
    	Misc.verboseln("Generalized Rabin Acceptance Condition w/ fixed order");
    	Misc.verboseln(""+condition);
    	
    	// initial state
    	//Misc.verboseln("Number of copies of DFA: " + condition.size());
    	List<Integer> iSCoords = new ArrayList<Integer>(condition.size());
		for (int i = 0; i < condition.size(); i++) {
    		iSCoords.add(i, 0);
    	}
    	initialState = new Tuple<Tuple<Formula, KState>, List<Integer>>(dfaA.initialState, iSCoords);

        states = new HashSet<Tuple<Tuple<Formula, KState>, List<Integer>>>();
        statesToNumbers= new HashMap<Tuple<Tuple<Formula, KState>, List<Integer>>,Integer>();
	currentNumber=0;
        states.add(initialState);
        statesToNumbers.put(initialState,currentNumber);
	currentNumber++;
        edges = new HashMap<Tuple<Tuple<Tuple<Formula, KState>,List<Integer>>, Tuple<Tuple<Formula, KState>,List<Integer>>>, ValuationSet>();
        Stack<Tuple<Tuple<Formula, KState>, List<Integer>>> workstack = new Stack<Tuple<Tuple<Formula, KState>, List<Integer>>>();
        workstack.push(initialState);

        int loopCount = 0;
        while (!workstack.empty()) {
        
        	Misc.verboseln("loop count: " + loopCount++);
        	Misc.verboseln("workstack : " + workstack);
        	Misc.verboseln("workcount : " + workstack.size());
        	
        	Tuple<Tuple<Formula, KState>, List<Integer>> curr = workstack.pop();
            Misc.verboseln("Curr: " + curr);
            for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : dfaA.edges.entrySet()) {
            	if (edge.getKey().left.equals(curr.left)) {
	                Misc.verboseln("\tKDFA A edge: " + edge);
	                // successor state
	                
	                
	                	Map<Valuation,List<Integer>> coords = new HashMap<Valuation,List<Integer>>(edge.getValue().size());
	                		for (Valuation v : edge.getValue())
	                				coords.put(v,curr.right);
	                		
	                		Misc.verboseln("init"+coords);
	            
	                	for (int i = 0; i < condition.size(); i++) {
	                		int mi = curr.right.get(i);	
	                		
	                		for (Valuation v : edge.getValue()){
	                			
	                			Map<Tuple<Tuple<Formula, KState>,Tuple<Formula, KState>>,ValuationSet> reachEventuallyERF = condition.get(i).right.get(mi);
	                			int m=mi;
	                			//Misc.verboseln("probihapred"+v+coords);
	                			//int c=0;
	                			while((reachEventuallyERF.containsKey(edge.getKey())) && (reachEventuallyERF.get(edge.getKey()).contains(v)) && ((m+1)<condition.get(i).right.size())){ 
	                						m++;
	                						//c++;
	                						//Misc.verboseln("probiha1"+v+coords);
	                						List<Integer> currList= new ArrayList<Integer>(coords.get(v));
	                						currList.set(i, m);
	                						coords.put(v, currList);
	                						reachEventuallyERF = condition.get(i).right.get(m);
	                						//Misc.verboseln("probiha2"+v+coords);
	                			}
	                			if((reachEventuallyERF.containsKey(edge.getKey())) && (reachEventuallyERF.get(edge.getKey()).contains(v)) && ((mi+1)==condition.get(i).right.size())){ 
	                				List<Integer> currList= new ArrayList<Integer>(coords.get(v));
	                				currList.set(i, 0);
	            					coords.put(v, currList);
	            					
	                			}
	                			
	                		}
	                		
	                	}
	                	 
	                	for (Valuation v : edge.getValue()){
	                		Tuple<Tuple<Formula, KState>, List<Integer>> state2= new Tuple<Tuple<Formula, KState>, List<Integer>>(edge.getKey().right,coords.get(v));
	                		Tuple<Tuple<Tuple<Formula, KState>, List<Integer>>, Tuple<Tuple<Formula, KState>, List<Integer>>> st = new Tuple<Tuple<Tuple<Formula, KState>, List<Integer>>, Tuple<Tuple<Formula, KState>, List<Integer>>>(curr,state2);
	                		if (!states.contains(state2)) {
                                            Misc.verboseln("\tNew state");
                                            states.add(state2);
                                            statesToNumbers.put(state2,currentNumber);
                                            currentNumber++;
                                            workstack.push(state2);
                                        }
	                		
	                		if(edges.containsKey(st)){
	                			ValuationSet vs = edges.get(st);
	                			vs.add(v);
	                			edges.put(st, vs);
	                		}
	                		else{
	                			ValuationSet vs = new ValuationSet();
	                			vs.add(v);
	                			edges.put(st, vs);
	                		}
	                		Misc.verboseln("edge: "+curr+"   "+v+"   "+coords.get(v));	                		
	                		
	                	}
	                	
	                	
	                	//return new Tuple<Tuple<Formula, KState>, List<Integer>>(edge.right, coords);
	                }

            	}
            }
        Misc.verboseln("Initial state: " + initialState);
        Misc.verboseln("Edges:         " + edges);
    }
    
    
    
    public String toDotty() {
		String dot = "digraph \"GenRabinAutomaton " + initialState +"\" {\n";
		//        Misc.verboseln("States       : " + states);
		//        Misc.verboseln("# states     : " + states.size());
		// The initial state in the set states may be different from the states in the set of edges. This can happen
		// because HashSet does not guarantee that two segment states that are equal have the same string representation.
		// To circumvent the problem, use the same string representation for rendering nodes and edges in the dotty file.
		Map<Tuple<Tuple<Formula, KState>, List<Integer>>, Tuple<Tuple<Formula, KState>, List<Integer>>> normalStates = new HashMap<Tuple<Tuple<Formula, KState>, List<Integer>>, Tuple<Tuple<Formula, KState>, List<Integer>>>();        
		for (Tuple<Tuple<Formula, KState>,List<Integer>> s : states) {
			normalStates.put(s,s);
			Misc.verbose(",");
			Formula display = displayLabels.get(s.left.left);

			KState k = s.left.right;
			String display2 = "";
			for(SegmentAutomaton sa : k.state.keySet()){
				//List<Formula> fs = (List<Formula>) k.state.get(sa).keySet();
				display2 += "[";
				for(Formula f : k.state.get(sa).keySet())
					display2 += "" + sa.displayLabels.get(f) + ":" + k.state.get(sa).get(f) + " ";
				display2 += "] ";
			}
			String display3 = s.right.toString();
			if (s.equals(initialState)) {
				dot += "\nnode [shape=oval, label=\"" + display + " :: " + display2 + "::" + display3 + "\"]\"" + s + "\";\n";
			} else {
				dot += "\nnode [shape=rectangle, label=\"" + display + " :: " + display2 + "::" + display3 + "\"]\"" + s + "\";\n";
			}
		}
		for (Entry<Tuple<Tuple<Tuple<Formula, KState>, List<Integer>>, Tuple<Tuple<Formula, KState>, List<Integer>>>, ValuationSet> edge : edges.entrySet()) {
			Misc.verbose(".");
			dot += "\n\"" + normalStates.get(edge.getKey().left) + "\" -> \"" + normalStates.get(edge.getKey().right) + "\" [label=\"" + BDDService.BDDtoString(BDDService.valuationSetToBDD(edge.getValue())) + "\"];\n"; 
		}
		return dot + "}\n";
	}
    
    
    public String toHOAF(KRabinAcceptanceCondition acc) {
                String dot = "";
                dot += "HOA: v1\n";
                dot += "tool: \"Rabinizer\" \"3\"\n";
                dot += "name: \"DTRA for "+ displayLabels.get(initialState.left.left)+"\"\n";
                dot += "properties: deterministic\n";
                dot += "States: " + states.size() +"\n";
		dot += "Start: " + statesToNumbers.get(initialState) +"\n";
            //    dot += "acc-name: Rabin\n";
		dot += "Acceptance: " + acc.toNumericalString(this) +"\n"; //TODO: handle trivial sets
		dot += "AP: "+Misc.bijectionIdAtom.size();
		for(Integer i=0;i<Misc.bijectionIdAtom.size();i++)
                    dot += " \"" + Misc.bijectionIdAtom.atom(i)+"\"";
		dot+= "\n";
                dot += "--BODY--\n";
		 
		// Map<Tuple<Formula, KState>, Tuple<Formula, KState>> normalStates = new HashMap<Tuple<Formula, KState>, Tuple<Formula, KState>>();        
		for (Tuple<Tuple<Formula, KState>, List<Integer>> s:states) {
			dot += "State: "+ statesToNumbers.get(s)+" \""+ s + "\"\n";
			
                        List<String> atoms= new ArrayList<String>();
			for(Integer i=Misc.bijectionIdAtom.size()-1;i>=0;i--){
				atoms.add((Misc.bijectionIdAtom.size()-1-i), Misc.bijectionIdAtom.atom(i));
			}
			str = "";
			iterateValuations(atoms,0,atoms.size(), s, acc);

			dot += str;
		}
		return dot + "--END--";
        }
    
    private void iterateValuations(List<String> atoms, int pos, int count, Tuple<Tuple<Formula, KState>, List<Integer>> s,KRabinAcceptanceCondition condition) {
		if (count <= 0) {
			Valuation val= new Valuation();
			for(String a:atoms)
				val.put(Misc.bijectionIdAtom.id(a),true);

			for(Integer i : Misc.bijectionIdAtom.idToAtom.keySet()){
				if(!val.containsKey(i)) val.put(i,false);
			}
			for(Map.Entry<Tuple<Tuple<Tuple<Formula, KState>, List<Integer>>, Tuple<Tuple<Formula, KState>, List<Integer>>>, ValuationSet> edge : edges.entrySet()){
				if(edge.getKey().left.equals(s)){
					if(edge.getValue().contains(val)){
						str+=" "+statesToNumbers.get(edge.getKey().right)+" ";
						int count2=0;
						for(Map<Tuple<Tuple<Tuple<Formula, KState>, List<Integer>>, Tuple<Tuple<Formula, KState>, List<Integer>>>, ValuationSet> edgeC: condition.edgesToInt.keySet()){
							Misc.verboseln("hrany"+edgeC+"hrana, co zkousim" + edge);
							if(edgeC.keySet().contains(edge.getKey())&&edgeC.get(edge.getKey()).contains(val)){
								count2++;
								if (count2==1)str+="{";
								if (count2>1)str+=" ";
								str+=condition.edgesToInt.get(edgeC);
							}
						}
						if(count2>0){str+="}";}
						str+="   \t/*" + val + "*/\n";
					}
				}
			}

		} else {
			// remove element in position pos 
			List<String> currPosRemoved = new ArrayList<String>(atoms.size());
			currPosRemoved.addAll(atoms);
			currPosRemoved.remove(pos);
			iterateValuations(currPosRemoved, pos, count - 1, s,condition);
			// keep element in position pos
			iterateValuations(atoms, pos + 1, count - 1, s,condition);



		}
	}
}
