package rabinizer.exec;

/* 
 * @ author Jan Kretinsky 
 * */ 

import java.util.*;
import java.util.Map.Entry;
import java.util.Arrays;

import rabinizer.formulas.*;

public class KState {
	
	public Map<SegmentAutomaton,Map<Formula,Integer>> state;
	public Map<Formula,SegmentAutomaton> segmentFormulasToSA;
	

	
	public KState(List<SegmentAutomaton> sAs) {
		state = new HashMap<SegmentAutomaton,Map<Formula,Integer>>();
		segmentFormulasToSA= new HashMap<Formula,SegmentAutomaton>(state.keySet().size());
		for(SegmentAutomaton sA : sAs){
			state.put(sA, new HashMap<Formula,Integer>());
			segmentFormulasToSA.put(sA.initialState,sA);
		}
	}
	
	public KState initialize(){
		for(SegmentAutomaton sA : state.keySet())
			state.get(sA).put(sA.initialState,1);
		return this;
	}
	
	public KState initializeFilled(){
		KState k = this.initialize();
		//for(Valuation v : Misc.valuations) // TODO rather number of states than valuations
		Valuation v = Misc.valuations.get(0);
		int max = 0;
		for(SegmentAutomaton sa : state.keySet())
			if(max < sa.states.size())
				max = sa.states.size();
		for(int i = 0; i < max; i++)
			k = k.step(state.keySet(),v); 
		this.state = k.state;
		return this;
	}
	
	public KState step(Set<SegmentAutomaton> relevantSAs, Valuation transitionValuation) {
		Misc.verboseln("currSegmentState :\n" + this);
		List<SegmentAutomaton> nextSAs = new ArrayList<SegmentAutomaton>(relevantSAs);
		//for(SegmentAutomaton sa : state.keySet())
		//	if(relevantSAs.contains(sa))
		//		nextSAs.add(sa);
		KState nextState = new KState(nextSAs);
		for(SegmentAutomaton sa : nextSAs){
			Map<Formula,Integer> config = new HashMap<Formula,Integer>();
			if(state.keySet().contains(sa))
				// move tokens
				for(Formula curr : state.get(sa).keySet()){
					for (Entry<Tuple<Formula, Formula>, ValuationSet> edge : sa.edges.entrySet()) {
						Formula from = edge.getKey().left;
						Formula to = edge.getKey().right;
						ValuationSet label = edge.getValue();
						if(from.equals(curr) && label.contains(transitionValuation) && !sa.finalStates.contains(to)) {
							int newInt = state.get(sa).get(from);
							// keep only the lowest token
							if(config.containsKey(to))
								newInt = (config.get(to) < newInt ? config.get(to) : newInt);
							config.put(to, newInt);
						}
					}
				}
			else {
				List<SegmentAutomaton> single = new ArrayList<SegmentAutomaton>();
				single.add(sa);
				config = (new KState(single)).initializeFilled().state.get(sa);
				//System.out.println("%%%%%%%%" + sa.initialState + " was irrelevant in \n" + this);
			}
				
			Misc.verboseln(config.toString());
			
			//TODO recompute tokens
			int[] tokens = new int[config.keySet().size()];
			int i = 0;
			for(Formula curr : config.keySet()){
				tokens[i] = config.get(curr);
				i ++;
				//System.out.println("old:"+curr + state.get(sa).get(curr));
				//System.out.println("new:"+transitionValuation+curr + tokens[i-1]);
			}
			//for(int j=0;j<tokens.length;j++)
			//	System.out.println(tokens[j]+"o");
			Arrays.sort(tokens);
			//for(int j=0;j<tokens.length;j++)
			//	System.out.println(tokens[j]+"n");
			for(Formula curr : config.keySet()){
				for(int j=0;j<tokens.length;j++)
					if(config.get(curr).equals(tokens[j])){
						config.put(curr, j + 1);
					//	System.out.println("search:"+config.get(curr)+j);
					}						
			}
			
			//TODO add token to the initial state
			if(! config.containsKey(sa.initialState)){
				config.put(sa.initialState, config.keySet().size() + 1);
			//System.out.println("init:"+config.keySet().size()+" "+config.get(sa.initialState)+" "+ tokens.length);	
			}//else{System.out.println("noninit"+config.get(sa.initialState) + tokens.length);}
			nextState.state.put(sa, config);
		}
		//System.out.println("----------");
		Misc.verbose("succSegmentState :\n" + nextState);
		return nextState;
	}
	
	public KState removeSA(SegmentAutomaton sa){
		this.state.remove(sa);
		this.segmentFormulasToSA.remove(sa.initialState);
		return this;
	}
	
	public String toString() {
		String s = new String("");
		for (SegmentAutomaton sa : state.keySet())
			s += sa.initialState + ":" + state.get(sa) + ""; // TODO use stateLabels!
		return s;
	}
	
	
	
	public int hashCode() {
		return state.hashCode();
	}

	public boolean equals(Object o) {
		if (!(o instanceof KState)) return false;
		else { 
			return ((KState)o).state.equals(this.state);
		}
	}

}
