package rabinizer.exec;

import java.util.*;
import net.sf.javabdd.*;
import rabinizer.formulas.*;

/**
 * Global state & pervasive methods.
 * 
 * @author Ruslan Ledesma-Garza
 *
 */
public class Misc {
	
	public static boolean verbose;
	
	public static void verboseln(String str) {
		if(verbose) { System.out.println(str); }
	}
	
	public static void verbose(String str) {
		if(verbose) { System.out.print(str); }
	}
	
	/**
	 * The BDD factory for the purpose of constructing canonical representations of formulas.
	 */
	public static BDDFactory bddFactory = BDDFactory.init("java", 2, 2);

	/**
	 * The map from boolean atoms to BDD variables for the purpose of constructing BDDs over boolean atoms.
	 * 
	 * Populated by Formula.bdd().
	 */
	public static BijectionBooleanAtomBddVar bijectionBooleanAtomBddVar = new BijectionBooleanAtomBddVar();
	
	/**
	 * Cache for the representative of a given bdd
	 */
	private static Map<BDD, Formula> bddToRepresentative= new HashMap<BDD, Formula>();
	
	/**
	 * Return the cached representative of a BDD. 
	 */
	public static Formula representativeOfBdd(BDD bdd, Formula representativeCandidate) {
		if (bddToRepresentative.containsKey(bdd))
			return bddToRepresentative.get(bdd);
		else {
			bddToRepresentative.put(bdd, representativeCandidate);
			return representativeCandidate;
		}
	}
	
	public static Formula formulaOfBdd(BDD b) {
		throw new Error("Misc.formulaOfBdd: NYI");
	}
	
	/**
	 * The bijection between atom identifiers and atoms for the purpose of
	 * 1. Consistently creating literals across parsing of different formulas (e.g. during unit testing) by mapping atoms to identifiers
	 * 2. Rendering valuations by mapping identifiers to atoms
	 * 
	 * Created and populated during parsing.
	 */
	public static BijectionIdAtom bijectionIdAtom = null;
		
	/**
	 * All the possible valuations.
	 * 
	 * Created and populated by initializeValuations.
	 */
	public static List<Valuation> valuations = null;
	
	/**
	 * Create the valuations corresponding to n variables and store them in valuations.
	 * The memory consumption is in the order of 2^n.
	 * 
	 * @param n
	 * @return 
	 */
	public static void initializeValuations(int n) {
		valuations = new ArrayList<Valuation>();
		boolean[] values = new boolean[n];
		enumerateValuations(values, n - 1);
	}

	private static void enumerateValuations(boolean[] values, int id) {
		if (id <= 0) {
			values[0] = false;
			valuations.add(new Valuation(values));
			values[0] = true;
			valuations.add(new Valuation(values));
		} else {
			values[id] = false;
			enumerateValuations(values, id - 1);
			values[id] = true;
			enumerateValuations(values, id - 1);
		}
	}

	public static BooleanConstant fTrue = new BooleanConstant(true);
    public static BooleanConstant fFalse = new BooleanConstant(false);
}
