package rabinizer.formulas;

import java.util.*;

import net.sf.javabdd.*;
import rabinizer.exec.*;

/**
 * @author Andreas Gaiser & Ruslan Ledesma-Garza
 *
 */
public class Conjunction extends Formula {

	public Formula left, right;

	public Conjunction(Formula f1, Formula f2) {
		this.left = f1;
		this.right = f2;
	}

	public BDD bdd() {
		if (cachedBdd == null) {
			cachedBdd = left.bdd().and(right.bdd());
			Misc.representativeOfBdd(cachedBdd, this);
		}
		return cachedBdd;
	}

	@Override
	public int hashCode() {
		return 3 * (left.hashCode() + right.hashCode());
	}

	@Override
	public boolean equals(Object o) {
		if (!(o instanceof Conjunction))
			return false;
		else			
			return ((Conjunction)o).left.equals(this.left) && ((Conjunction)o).right.equals(this.right);
	}

	@Override
	public Formula negated() {
		return new Disjunction(left.negated(), right.negated());
	}

	@Override
	public String toReversePolishString() {
		return "& " + left.toReversePolishString() + " " + right.toReversePolishString();
	}

	@Override
	public String toString() {
		if (strFormula == null) strFormula = "(" + left.toString() + " & " + right.toString() + ")";
		return strFormula;
	}

	@Override
	public Tuple<Set<Formula>, Set<Formula>> recurrentFormulas(boolean accumulate) {
		Tuple<Set<Formula>, Set<Formula>> l = left.recurrentFormulas(accumulate);
		Tuple<Set<Formula>, Set<Formula>> r = right.recurrentFormulas(accumulate);
		l.left.addAll(r.left);
		l.right.addAll(r.right);
		return l;
	}

	@Override
	public Formula evaluateCurrentAssertions(Valuation valuation) {
		return new Conjunction(left.evaluateCurrentAssertions(valuation),
				right.evaluateCurrentAssertions(valuation));	
	}

	@Override
	public Formula removeConstantCurrentAssertions() {
		Formula new_left = left.removeConstantCurrentAssertions();
		if (new_left instanceof BooleanConstant) {
			if (((BooleanConstant)new_left).value) {
				return right.removeConstantCurrentAssertions(); 
			} else {
				return new BooleanConstant(false);
			}
		} else { 
			Formula new_right = right.removeConstantCurrentAssertions();
			if (new_right instanceof BooleanConstant) {
				if (((BooleanConstant)new_right).value){
					return new_left;
				} else {
					return new BooleanConstant(false);
				}
			} else {
				return new Conjunction(new_left, new_right);
			}
		}
	}

	@Override
	public Formula removeX() {
		return new Conjunction(left.removeX(), right.removeX());
	}

	@Override
	public Formula unfold() {
		return new Conjunction(left.unfold(), right.unfold());
	}

	@Override
	public Formula unfoldNoG() {
		return new Conjunction(left.unfoldNoG(), right.unfoldNoG());
	}

	@Override
	public boolean isProgressFormula() {
		return left.isProgressFormula() || right.isProgressFormula();
	}

	@Override
	public boolean hasSuspendableG(Formula f) {
		//System.out.println("$$$"+left+"\n$$$"+right);
		/*
		if (right instanceof XOperator) {
			Formula g = ((XOperator)right).operand;
			if (g instanceof GOperator){
				//System.out.println("$$$$$" + ((GOperator)g).operand.unfold() + "$$$$" + left);
				//GOperator g = (GOperator) right;
				if (((GOperator)g).operand.unfold().hasSubformula(left)){
					System.out.println("no");
					return false;			// FIXME instantiation of the unfold
				}
			}
		}*/

		if (! left.isVeryDifferentFrom(right)){
			//System.out.println("$$$$$$not very different$$$$"+this);
			return false;
		}
		if (left.equals(f.unfold()) && right.isUnfoldOfF() && !right.containsG() ||
				right.equals(f.unfold()) && left.isUnfoldOfF() && !left.containsG()){		
			//System.out.println("$$$$$$"+left+"$$$$"+right+"$$$susp "+f);
			return true;					// independent waiting formula
		}
		else
			return left.hasSuspendableG(f) && right.hasSuspendableG(f); 	// don't know yet
		
		/* NO OPTIMIZATION version
		if (! left.isVeryDifferentFrom(right))
			return false;
		if (left.equals(f.unfold()) && right.isUnfoldOfF() ||
				right.equals(f.unfold()) && left.isUnfoldOfF() )		
			return false; // TODO FIXME true;					// independent waiting formula
		else
			return left.hasSuspendableG(f) && right.hasSuspendableG(f); 	// don't know yet
		*/
	}
	
	@Override
	public boolean isVeryDifferentFrom(Formula f) {
		return left.isVeryDifferentFrom(f) || right.isVeryDifferentFrom(f);
	}
	
	@Override
	public boolean containsG() {
		return left.containsG() || right.containsG();
	}

	@Override
	public boolean hasSubformula(Formula f) {
		return this.equals(f) || left.hasSubformula(f) || right.hasSubformula(f);
	}

	@Override
	public Formula removeXsFromCurrentBooleanAtoms() {
		return new Conjunction(left.removeXsFromCurrentBooleanAtoms(), right.removeXsFromCurrentBooleanAtoms());
	}

	@Override
	public Set<Formula> argumentsAlways() {
		Set<Formula> r = left.argumentsAlways();
		r.addAll(right.argumentsAlways());
		return r;
	}

	@Override
	public Set<Formula> argumentsFinsideG(boolean acc) {
		Set<Formula> r = left.argumentsFinsideG(acc);
		r.addAll(right.argumentsFinsideG(acc));
		return r;
	}

	public boolean untilOcurrs() {
		return left.untilOcurrs() || right.untilOcurrs();
	}

	public Formula toNNF() {
		return new Conjunction(left.toNNF(), right.toNNF());
	}
}
