package tests;

import static org.junit.Assert.*;

import java.io.StringReader;
import java.util.Arrays;
import java.util.Collection;


import org.junit.Test;

import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import rabinizer.formulas.Formula;
import rabinizer.parser.LTLParser;

@RunWith(value = Parameterized.class)
public class Parsing {

	private String formula;
	private boolean expected;
	
	public Parsing(String formula, boolean ok) {
		this.formula = formula;
		this.expected = ok;
	}

	@Parameters
	public static Collection<Object[]> data() {
		// Test cases, valid and invalid formulae
		Object[][] data = new Object[][] { 
				{ "", false},
				{ "a1", true }, 
				{ "a2aopore !& biIII12", false },
				{ "!AA123", true}, 
				{ "!aAA&!bii1&!c2213     &!d113", true },
				{ "F 111ai & a & !i & G i", true },
				{ "F a & G b & G (a & b &)", false },
				{ " F F F", false },
				{ "F G F G F G f & g", true },
				{ "X a", true },
				{ "a U b", true },
				{ "F ! a", true },
				{ "F (! a)", true },
				{ "F (a & b)", true },
				{ "(a | b) & c", true },
				{ "F (a U b)", true },
				{ "G (a U b)", true },
			};
		return Arrays.asList(data);
	}

	@Test
	public void pushTest() {
		System.out.println("TEST CASE : "+this.getClass().getCanonicalName()+"("+formula+")");
		
		LTLParser parser = new LTLParser(new StringReader(formula));
		Formula f = null;
		try { f = parser.parse(); }
		catch (Exception e) { assertFalse(expected); }
		
		assertTrue((f != null) == expected);
	}

}
