package tests;

import static org.junit.Assert.*;

import java.io.StringReader;
import java.util.Arrays;
import java.util.Collection;

import org.junit.Test;

import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import rabinizer.formulas.Formula;
import rabinizer.parser.LTLParser;

@RunWith(value = Parameterized.class)
public class Unfold {

	private String input;
	private String expected;	
	
	public Unfold(String input, String expected) {
		this.input = input;
		this.expected = expected;
	}

	@Parameters
	public static Collection<Object[]> data() {
		Object[][] data = new Object[][] {
				{ "X a", "X a" },
				{ "G a", "a & X G a" },
				{ "F a", "a | X F a" },
				{ "a U b", "b | a & X (a U b)" },

				{ "a U X b", "X b | (a & X (a U X b))" },
				{ "G a & b", "a & X G a & b" },
				{ "b & G a", "b & (a & X G a)" },
				{ "b | G a", "b | a & X G a" },
				
				{ "G F a | G F b", "(a | X F a) & X G F a | (b | X F b) & X G F b" },
				{ "(a | X F a) & X G F a | (b | X F b) & X G F b", "(a | X F a) & X G F a | (b | X F b) & X G F b" },
		};
		return Arrays.asList(data);
	}

	@Test
	public void pushTest() {
		System.out.println(this.getClass().getCanonicalName()+" : "+input+" and "+expected);
		
		LTLParser parser = new LTLParser(new StringReader(input));
		Formula inf = null;
		try { inf = parser.parse(); }
		catch (Exception e) { assertTrue(false); }
		
		parser.ReInit(new StringReader(expected));
		Formula exf = null;
		try { exf = parser.parsePreviousAtoms(); }
		catch (Exception e) { assertTrue(false); }
		
		System.out.println("expected : " + exf);
		System.out.println("actual   : " + inf.unfold());
		assertTrue(exf.equals(inf.unfold()));
	}

}
